/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *      $RCSfile: RayShadeDisplayDevice.C,v $
 *      $Author: dalke $        $Locker:  $                $State: Exp $
 *      $Revision: 1.5 $      $Date: 96/03/24 07:07:20 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 * FileRenderer type for the Rayshade raytracer 
 *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "RayShadeDisplayDevice.h"
#include "Stack.h"
#include "Matrix4.h"
#include "DispCmds.h"
#include "Inform.h"
#include "utilities.h"

#define DEFAULT_RADIUS 0.002
#define DASH_LENGTH 0.02
#define SCALE_FACTOR 1.3

///////////////////////// constructor and destructor

// constructor ... initialize some variables
RayShadeDisplayDevice::RayShadeDisplayDevice(char *nm) : FileRenderer(nm) {
	
    MSGDEBUG(1,"Creating RayShadeDisplayDevice ..." << sendmsg);
    
    strcpy(renderCommand,"rayshade < %s > %s.rle");
    strcpy(fname,"plot.ray");
}
        
// destructor
RayShadeDisplayDevice::~RayShadeDisplayDevice(void) { }

///////////////////////// protected nonvirtual routines

// fix the scaling
float RayShadeDisplayDevice::scale_fix(float x) {
    return ( x / SCALE_FACTOR );  
}

// draw a point
void RayShadeDisplayDevice::point(float spdata[]) {
    float vec[3];

    // transform the world coordinates
    (transMat.top()).multpoint3d(spdata, vec);
   
    // draw the sphere

    fprintf(ray_file, "applysurf diffuse %3.2f %3.2f %3.2f specular 1.0 1.0 1.0 ", 
        currentColor[0], currentColor[1], currentColor[2]);
    fprintf(ray_file, "specpow 40.0\n");
    fprintf(ray_file, "sphere %5f %5f %5f %5f\n", 
    scale_fix(float(lineWidth)*DEFAULT_RADIUS),
        scale_fix(vec[0]), scale_fix(vec[1]), scale_fix(vec[2]));
 }

// draw a sphere
void RayShadeDisplayDevice::sphere(float spdata[]) {
    float vec[3];
    float radius;
    
    // transform the world coordinates
    (transMat.top()).multpoint3d(spdata, vec);
    radius = scale_radius(spdata[3]);
   
    // draw the sphere

    fprintf(ray_file, "applysurf diffuse %3.2f %3.2f %3.2f specular 1.0 1.0 1.0 ", 
        currentColor[0], currentColor[1], currentColor[2]);
    fprintf(ray_file, "specpow 40.0\n");
    fprintf(ray_file, "sphere %5f %5f %5f %5f\n", scale_fix(radius), 
        scale_fix(vec[0]), scale_fix(vec[1]), scale_fix(vec[2]));
}

// draw a line (cylinder) from a to b
void RayShadeDisplayDevice::line(float *a, float*b) {
    int i, j, test;
    float dirvec[3], unitdirvec[3];
    float from[3], to[3], tmp1[3], tmp2[3];
    float len;
    
    if(lineStyle == ::SOLIDLINE ) {
  
        // transform the world coordinates
        (transMat.top()).multpoint3d(a, from);
        (transMat.top()).multpoint3d(b, to);
    
        // draw the cylinder
        
        fprintf(ray_file, "applysurf diffuse %3.2f %3.2f %3.2f specular 1.0 1.0 1.0 ", 
            currentColor[0], currentColor[1], currentColor[2]);
        fprintf(ray_file, "specpow 40.0\n");
        fprintf(ray_file, "cylinder %5f %5f %5f %5f %5f %5f %5f\n",
            scale_fix(float(lineWidth)*DEFAULT_RADIUS),
            scale_fix(from[0]), scale_fix(from[1]), scale_fix(from[2]),
            scale_fix(to[0]), scale_fix(to[1]), scale_fix(to[2])); 
        
    } else if (lineStyle == ::DASHEDLINE ) {
        
         // transform the world coordinates
        (transMat.top()).multpoint3d(a, tmp1);
        (transMat.top()).multpoint3d(b, tmp2);

        // how to create a dashed line
        for(i=0;i<3;i++) {
            dirvec[i] = tmp2[i] - tmp1[i];  // vector from a to b
        }
        len = sqrtf( dirvec[0]*dirvec[0] + dirvec[1]*dirvec[1] + dirvec[2]*dirvec[2] );
        for(i=0;i<3;i++) {
            unitdirvec[i] = dirvec[i] / sqrtf(len);  // unit vector pointing from a to b
        }
           
        test = 1;
        i = 0;

        while( test == 1 ) {
            for(j=0;j<3;j++) {
                from[j] = tmp1[j] + (2*i)*DASH_LENGTH*unitdirvec[j];
                to[j] = tmp1[j] + (2*i + 1)*DASH_LENGTH*unitdirvec[j];
            }
            if( fabsf(tmp1[0] - to[0]) >= fabsf(dirvec[0]) ) {
                for(j=0;j<3;j++) {
                    to[j] = tmp2[j];
                }
                test = 0;
            }
    
            // draw the cylinder

            fprintf(ray_file, "applysurf diffuse %3.2f %3.2f %3.2f specular 1.0 1.0 1.0 ", 
                currentColor[0], currentColor[1], currentColor[2]);
            fprintf(ray_file, "specpow 40.0\n");
            fprintf(ray_file, "cylinder %5f %5f %5f %5f %5f %5f %5f\n",
                scale_fix(float(lineWidth)*DEFAULT_RADIUS),
                scale_fix(from[0]), scale_fix(from[1]), scale_fix(from[2]),
                scale_fix(to[0]), scale_fix(to[1]), scale_fix(to[2])); 
            
            i++;
        }
    } else {
        msgErr << "RayShadeDisplayDevice: Unknown line style " << lineStyle << sendmsg;
    }

}

// draw a cylinder
void RayShadeDisplayDevice::cylinder(float *a, float *b, float r) {
  float from[3], to[3];
  float radius;
  
  // transform the world coordinates
  (transMat.top()).multpoint3d(a, from);
  (transMat.top()).multpoint3d(b, to);
  radius = scale_radius(r);
    
  // draw the cylinder
  
        // draw the cylinder

        fprintf(ray_file, "applysurf diffuse %3.2f %3.2f %3.2f specular 1.0 1.0 1.0 ", 
            currentColor[0], currentColor[1], currentColor[2]);
        fprintf(ray_file, "specpow 40.0\n");
        fprintf(ray_file, "cylinder %5f %5f %5f %5f %5f %5f %5f\n",
            scale_fix(radius),
            scale_fix(from[0]), scale_fix(from[1]), scale_fix(from[2]),
            scale_fix(to[0]), scale_fix(to[1]), scale_fix(to[2])); 

	// put disks on the ends
  fprintf(ray_file, "disc %5f %5f %5f %5f %5f %5f %5f\n",
	  scale_fix(radius),
	  scale_fix(from[0]), scale_fix(from[1]), scale_fix(from[2]),
	  scale_fix(from[0]-to[0]), scale_fix(from[1]-to[1]), 
	  scale_fix(from[2]-to[2]));
  fprintf(ray_file, "disc %5f %5f %5f %5f %5f %5f %5f\n",
	  scale_fix(radius),
	  scale_fix(to[0]), scale_fix(to[1]), scale_fix(to[2]),
	  scale_fix(to[0]-from[0]), scale_fix(to[1]-from[1]), 
	  scale_fix(to[2]-from[2]));


}

// draw a cone
void RayShadeDisplayDevice::cone(float *a, float *b, float r) {
  float from[3], to[3];
  float radius;
  
  // transform the world coordinates
  (transMat.top()).multpoint3d(a, from);
  (transMat.top()).multpoint3d(b, to);
  radius = scale_radius(r);
    
    
  // draw the cone

    fprintf(ray_file, "applysurf diffuse %3.2f %3.2f %3.2f specular 1.0 1.0 1.0 ", 
        currentColor[0], currentColor[1], currentColor[2]);
    fprintf(ray_file, "specpow 40.0\n");
    fprintf(ray_file, "cone %5f %5f %5f %5f %5f %5f %5f %5f\n",
        scale_fix(radius), 
        scale_fix(from[0]), scale_fix(from[1]), scale_fix(from[2]), 
        scale_fix(float(lineWidth)*DEFAULT_RADIUS), 
        scale_fix(to[0]), scale_fix(to[1]), scale_fix(to[2]));
}

// draw a triangle
void RayShadeDisplayDevice::triangle(float *a, float *b, float *c, float *n1, float *n2, float *n3) {
    float vec1[3], vec2[3], vec3[3];
    float norm1[3], norm2[3], norm3[3];
  
  
    // transform the world coordinates
    (transMat.top()).multpoint3d(a, vec1);
    (transMat.top()).multpoint3d(b, vec2);
    (transMat.top()).multpoint3d(c, vec3);
    (transMat.top()).multpoint3d(n1, norm1);
    (transMat.top()).multpoint3d(n2, norm2);
    (transMat.top()).multpoint3d(n3, norm3);

    // draw the triangle

    fprintf(ray_file, "applysurf diffuse %3.2f %3.2f %3.2f specular 1.0 1.0 1.0 ", 
        currentColor[0], currentColor[1], currentColor[2]);
    fprintf(ray_file, "specpow 40.0\n");
    fprintf(ray_file, "triangle %5f %5f %5f %5f %5f %5f ", 
        scale_fix(vec1[0]), scale_fix(vec1[1]), scale_fix(vec1[2]),
        scale_fix(norm1[0]), scale_fix(norm1[1]), scale_fix(norm1[2]));
    fprintf(ray_file, "%5f %5f %5f %5f %5f %5f ",
        scale_fix(vec2[0]), scale_fix(vec2[1]), scale_fix(vec2[2]),
        scale_fix(norm2[0]), scale_fix(norm2[1]), scale_fix(norm2[2]));
    fprintf(ray_file, "%5f %5f %5f %5f %5f %5f\n",
        scale_fix(vec3[0]), scale_fix(vec3[1]), scale_fix(vec3[2]),
        scale_fix(norm3[0]), scale_fix(norm3[1]), scale_fix(norm3[2]));
  
}

// draw a square
void RayShadeDisplayDevice::square(float *norm, float *a, float *b, float *c, float *d) {
  
     // draw as two triangles
     triangle(a, b, c, norm, norm, norm);
     triangle(a, c, d, norm, norm, norm);
  
}

///////////////////// public virtual routines

// initialize the file for output
int RayShadeDisplayDevice::prepare3D(int bar) {
    time_t t;
    
    bar = bar;

    if(!(ray_file = fopen(fname,"w"))) {
        msgErr << "Cannot open file " << fname << " for Raster3d output" << sendmsg;
        Initialized = FALSE;
    } else {
  
        Initialized = TRUE;
        ray_filename = stringdup(fname);
    
        // file for RayShade raytracer.
        
        t = time(NULL);
        fprintf(ray_file,"/* Rayshade input script: %s\n", fname);
        fprintf(ray_file,"   Generated for user: %s\n",cuserid(NULL));
        fprintf(ray_file,"   Creation date: %s",asctime(localtime(&t)));
        fprintf(ray_file,"  ---------------------------- */\n\n");
        
                
        // The current view 
        
        fprintf(ray_file, "\n/* Define current view */\n");        
        fprintf(ray_file, "eyep %5f %5f %5f\n", eyePos[0], eyePos[1], eyePos[2]);
        fprintf(ray_file, "lookp %5f %5f %5f\n", eyeDir[0], eyeDir[1], eyeDir[2]);
        fprintf(ray_file, "up %5f %5f %5f\n", 0.0, 1.0, 0,0);
        fprintf(ray_file, "fov %5f\n", 45.0);
        if(stereo_mode()) {
            fprintf(ray_file, "eyesep %5f\n", eyesep() );
        }
        fprintf(ray_file, "maxdepth 10\n");
        
        // Size of image in pixels
        fprintf(ray_file, "screen %d %d\n", xSize, ySize);
      
        // Lights
        
        fprintf(ray_file, "\n/* Light Definitions */\n");
        for(int i=0;i<DISP_LIGHTS;i++) {
            if(lightDefined[i] && lightOn[i]) {
                fprintf(ray_file, "light %3.2f %3.2f %3.2f ", lightColor[i][0],
                    lightColor[i][1], lightColor[i][2]);
                fprintf(ray_file, "directional %5f %5f %5f\n", lightPos[i][0],
                    lightPos[i][1], lightPos[i][2]);
           }
       }
       
       // background color
       fprintf(ray_file, "\n/* Set background color */\n");
       fprintf(ray_file, "background %3.2f %3.2f %3.2f\n", 
           backColor[0], backColor[1], backColor[2]);
                  
       // start the objects
       fprintf(ray_file, "\n/* Start object descriptions */\n");       
    }    
    // that's all for the header.
    return Initialized;
}

    
// clean up after yourself
void RayShadeDisplayDevice::update(int foo) {

  foo = foo;
  
  if(Initialized) {
    fprintf(ray_file,"\n/* End of File */\n");
    fclose(ray_file);
    if(ray_filename) delete [] ray_filename;
    msgInfo << "Rayshade file generation finished" << sendmsg;
  }
}
